from litescope.software.dump.common import DumpData, DumpVariable, Dump
from litescope.software.dump.csv import CSVDump
from litescope.software.dump.python import PythonDump
from litescope.software.dump.json import JSONDump
from litescope.software.dump.sigrok import SigrokDump
from litescope.software.dump.vcd import VCDDump



#
# This file is part of LiteScope.
#
# Copyright (c) 2021 Arnaud Durand <arnaud.durand@unifr.ch>
# SPDX-License-Identifier: BSD-2-Clause

import json

from litescope.software.dump.common import Dump


class JSONDump(Dump):
    def __init__(self, dump=None):
        Dump.__init__(self)
        self.variables = [] if dump is None else dump.variables

    def generate_data(self):
        return {v.name: v.values for v in self.variables}

    def write(self, filename):
        with open(filename, "w") as f:
            json.dump(self.generate_data(), f)

    def read(self, filename):
        raise NotImplementedError("JSON files can not (yet) be read, please contribute!")



#
# This file is part of LiteScope.
#
# Copyright (c) 2015-2017 Florent Kermarrec <florent@enjoy-digital.fr>
# SPDX-License-Identifier: BSD-2-Clause

import os
import math
import shutil
import zipfile
import re
from collections import OrderedDict

from litescope.software.dump.common import Dump, DumpVariable


class SigrokDump(Dump):
    def __init__(self, dump=None, samplerate=None):
        Dump.__init__(self)
        self.variables = [] if dump is None else dump.variables
        self.samplerate = 100e6 if samplerate is None else samplerate

    def write_version(self):
        f = open("version", "w")
        f.write("1")
        f.close()

    def write_metadata(self, name):
        probe_bits = math.ceil(sum(variable.width for variable in self.variables)/8)*8
        f = open("metadata", "w")
        r = """
[global]
sigrok version=0.3.0
[device 1]
capturefile=logic-1-1
total probes={}
samplerate={} MHz
unitsize={}
""".format(
        probe_bits,
        self.samplerate//1e6*2,
        probe_bits//8
    )
        i = 1
        for variable in self.variables:
            if variable.width > 1:
                for j in range(variable.width):
                    r += "probe{}={}[{}]\n".format(i, variable.name, j)
                    i += 1
            else:
                r += "probe{}={}\n".format(i, variable.name)
                i += 1
        f.write(r)
        f.close()

    def write_data(self):
        data_bits = math.ceil(sum(variable.width for variable in self.variables)/8)*8
        data_len = 0
        for variable in self.variables:
            data_len = max(data_len, len(variable))
        datas = []
        for i in range(data_len):
            data = 0
            for j, variable in enumerate(reversed(self.variables)):
                data = data << variable.width
                try:
                    data |= variable.values[i]
                except:
                    pass
            datas.append(data)
        f = open("logic-1-1", "wb")
        for data in datas:
            f.write(data.to_bytes(data_bits//8, "little"))
        f.close()

    def zip(self, name):
        f = zipfile.ZipFile(name + ".sr", "w")
        os.chdir(name)
        f.write("version")
        f.write("metadata")
        f.write("logic-1-1")
        os.chdir("..")
        f.close()

    def write(self, filename):
        name, ext = os.path.splitext(filename)
        if os.path.exists(name):
            shutil.rmtree(name)
        os.makedirs(name)
        os.chdir(name)
        self.write_version()
        self.write_metadata(name)
        self.write_data()
        os.chdir("..")
        self.zip(name)
        shutil.rmtree(name)

    def unzip(self, filename, name):
        f = open(filename, "rb")
        z = zipfile.ZipFile(f)
        if os.path.exists(name):
            shutil.rmtree(name)
            os.makedirs(name)
        for file in z.namelist():
            z.extract(file, name)
        f.close()

    def read_metadata(self):
        probes = OrderedDict()
        f = open("metadata", "r")
        for l in f:
            m = re.search("probe([0-9]+) = (\w+)", l, re.I)
            if m is not None:
                index = int(m.group(1))
                name = m.group(2)
                probes[name] = index
            m = re.search("samplerate = ([0-9]+) kHz", l, re.I)
            if m is not None:
                self.samplerate = int(m.group(1))*1000
            m = re.search("samplerate = ([0-9]+) MHz", l, re.I)
            if m is not None:
                self.samplerate = int(m.group(1))*1000000
        f.close()
        return probes

    def read_data(self, name, nprobes):
        datas = []
        f = open("logic-1-1", "rb")
        while True:
            data = f.read(math.ceil(nprobes/8))
            if data == bytes('', "utf-8"):
                break
            data = int.from_bytes(data, "little")
            datas.append(data)
        f.close()
        return datas

    def read(self, filename):
        self.variables = []
        name, ext = os.path.splitext(filename)
        self.unzip(filename, name)
        os.chdir(name)
        probes = self.read_metadata()
        datas = self.read_data(name, len(probes.keys()))
        os.chdir("..")
        shutil.rmtree(name)

        for k, v in probes.items():
            probe_data = []
            for data in datas:
                probe_data.append((data >> (v-1)) & 0x1)
            self.add(DumpVariable(k, 1, probe_data))



#
# This file is part of LiteScope.
#
# Copyright (c) 2015-2019 Florent Kermarrec <florent@enjoy-digital.fr>
# Copyright (c) 2019 kees.jongenburger <kees.jongenburger@gmail.com>
# SPDX-License-Identifier: BSD-2-Clause

def dec2bin(d, width=0):
    if d == "x":
        return "x"*width
    elif d == 0:
        b = "0"
    else:
        b = ""
        while d != 0:
            b = "01"[d&1] + b
            d = d >> 1
    return b.zfill(width)


def get_bits(values, low, high=None):
    r = []
    if high is None:
        high = low + 1
    for val in values:
        t = (val >> low) & (2**(high - low) - 1)
        r.append(t)
    return r


class DumpData(list):
    def __init__(self, width):
        self.width = width

    def __getitem__(self, key):
        if isinstance(key, int):
            return get_bits(self, key)
        elif isinstance(key, slice):
            if key.start != None:
                start = key.start
            else:
                start = 0
            if key.stop != None:
                stop = key.stop
            else:
                stop = self.width
            if stop > self.width:
                stop = self.width
            if key.step != None:
                raise KeyError
            return get_bits(self, start, stop)
        else:
            raise KeyError


class DumpVariable:
    def __init__(self, name, width, values=[]):
        self.name = name
        self.width = width
        self.values = [int(v)%2**width for v in values]

    def __len__(self):
        return len(self.values)


class Dump:
    def __init__(self):
        self.variables = []

    def add(self, variable):
        self.variables.append(variable)

    def add_from_layout(self, layout, variable):
        offset = 0
        for name, sample_width in layout:
            values = variable[offset:offset+sample_width]
            values2x = [values[i//2] for i in range(len(values)*2)]
            self.add(DumpVariable(name, sample_width, values2x))
            offset += sample_width

    def add_from_layout_flatten(self, layout, variable):
        offset = 0
        for name, sample_width in layout:
            # The samples from the logic analyzer end up in an array of size sample size
            # and have n (number of channel) bits. The following does a bit slice on the array
            # elements (implemented above)
            values         = variable[offset:offset+sample_width]
            values_flatten = [values[i//sample_width] >> (i % sample_width ) & 1 for i in range(len(values)*sample_width)]
            self.add(DumpVariable(name, 1, values_flatten))
            offset += sample_width

    def add_scope_clk(self):
        self.add(DumpVariable("scope_clk", 1, [1, 0]*(len(self)//2)))

    def add_scope_trig(self, offset):
        self.add(DumpVariable("scope_trig", 1, [0]*offset + [1]*(len(self)-offset)))

    def __len__(self):
        l = 0
        for variable in self.variables:
            l = max(len(variable), l)
        return l



#
# This file is part of LiteScope.
#
# Copyright (c) 2016-2019 Florent Kermarrec <florent@enjoy-digital.fr>
# Copyright (c) 2018 bunnie <bunnie@kosagi.com>
# Copyright (c) 2016 Tim 'mithro' Ansell <mithro@mithis.com>
# SPDX-License-Identifier: BSD-2-Clause

from migen import *
from migen.genlib.misc import WaitTimer
from migen.genlib.cdc import MultiReg, PulseSynchronizer

from litex.build.tools import write_to_file

from litex.soc.interconnect.csr import *
from litex.soc.cores.gpio import GPIOInOut
from litex.soc.interconnect import stream

# LiteScope IO -------------------------------------------------------------------------------------

class LiteScopeIO(Module, AutoCSR):
    def __init__(self, data_width):
        self.data_width = data_width
        self.input  = Signal(data_width)
        self.output = Signal(data_width)

        # # #

        self.submodules.gpio = GPIOInOut(self.input, self.output)

    def get_csrs(self):
        return self.gpio.get_csrs()

# LiteScope Analyzer -------------------------------------------------------------------------------

def core_layout(data_width):
    return [("data", data_width), ("hit", 1)]


class _Trigger(Module, AutoCSR):
    def __init__(self, data_width, depth=16):
        self.sink   = sink   = stream.Endpoint(core_layout(data_width))
        self.source = source = stream.Endpoint(core_layout(data_width))

        self.enable = CSRStorage()
        self.done   = CSRStatus()

        self.mem_write = CSR()
        self.mem_mask  = CSRStorage(data_width)
        self.mem_value = CSRStorage(data_width)
        self.mem_full  = CSRStatus()

        # # #

        # Control re-synchronization
        enable   = Signal()
        enable_d = Signal()
        self.specials += MultiReg(self.enable.storage, enable, "scope")
        self.sync.scope += enable_d.eq(enable)

        # Status re-synchronization
        done = Signal()
        self.specials += MultiReg(done, self.done.status)

        # Memory and configuration
        mem = stream.AsyncFIFO([("mask", data_width), ("value", data_width)], depth)
        mem = ClockDomainsRenamer({"write": "sys", "read": "scope"})(mem)
        self.submodules += mem
        self.comb += [
            mem.sink.valid.eq(self.mem_write.re),
            mem.sink.mask.eq(self.mem_mask.storage),
            mem.sink.value.eq(self.mem_value.storage),
            self.mem_full.status.eq(~mem.sink.ready)
        ]

        # Hit and memory read/flush
        hit   = Signal()
        flush = WaitTimer(2*depth)
        flush = ClockDomainsRenamer("scope")(flush)
        self.submodules += flush
        self.comb += [
            flush.wait.eq(~(~enable & enable_d)), # flush when disabling
            hit.eq((sink.data & mem.source.mask) == (mem.source.value & mem.source.mask)),
            mem.source.ready.eq((enable & hit) | ~flush.done),
        ]

        # Output
        self.comb += [
            sink.connect(source),
            # Done when all triggers have been consumed
            done.eq(~mem.source.valid),
            source.hit.eq(done)
        ]

class _SubSampler(Module, AutoCSR):
    def __init__(self, data_width):
        self.sink   = sink   = stream.Endpoint(core_layout(data_width))
        self.source = source = stream.Endpoint(core_layout(data_width))

        self.value = CSRStorage(16)

        # # #

        value = Signal(16)
        self.specials += MultiReg(self.value.storage, value, "scope")

        counter = Signal(16)
        done    = Signal()
        self.sync.scope += \
            If(source.ready,
                If(done,
                    counter.eq(0)
                ).Elif(sink.valid,
                    counter.eq(counter + 1)
                )
            )

        self.comb += [
            done.eq(counter == value),
            sink.connect(source, omit={"valid"}),
            source.valid.eq(sink.valid & done)
        ]


class _Mux(Module, AutoCSR):
    def __init__(self, data_width, n):
        self.sinks  = sinks  = [stream.Endpoint(core_layout(data_width)) for i in range(n)]
        self.source = source = stream.Endpoint(core_layout(data_width))

        self.value = CSRStorage(bits_for(n))

        # # #

        value = Signal(bits_for(n))
        self.specials += MultiReg(self.value.storage, value, "scope")

        cases = {}
        for i in range(n):
            cases[i] = sinks[i].connect(source)
        self.comb += Case(value, cases)


class _Storage(Module, AutoCSR):
    def __init__(self, data_width, depth):
        self.sink = sink = stream.Endpoint(core_layout(data_width))

        self.enable    = CSRStorage()
        self.done      = CSRStatus()

        self.length    = CSRStorage(bits_for(depth))
        self.offset    = CSRStorage(bits_for(depth))

        self.mem_valid = CSRStatus()
        self.mem_data  = CSRStatus(data_width)

        # # #

        # Control re-synchronization
        enable   = Signal()
        enable_d = Signal()
        self.specials += MultiReg(self.enable.storage, enable, "scope")
        self.sync.scope += enable_d.eq(enable)

        length = Signal().like(self.length.storage)
        offset = Signal().like(self.offset.storage)
        self.specials += MultiReg(self.length.storage, length, "scope")
        self.specials += MultiReg(self.offset.storage, offset, "scope")

        # Status re-synchronization
        done = Signal()
        self.specials += MultiReg(done, self.done.status)

        # Memory
        mem = stream.SyncFIFO([("data", data_width)], depth, buffered=True)
        mem = ClockDomainsRenamer("scope")(mem)
        cdc = stream.AsyncFIFO([("data", data_width)], 4)
        cdc = ClockDomainsRenamer(
            {"write": "scope", "read": "sys"})(cdc)
        self.submodules += mem, cdc

        # Flush
        mem_flush = WaitTimer(depth)
        mem_flush = ClockDomainsRenamer("scope")(mem_flush)
        self.submodules += mem_flush

        # FSM
        fsm = FSM(reset_state="IDLE")
        fsm = ClockDomainsRenamer("scope")(fsm)
        self.submodules += fsm
        fsm.act("IDLE",
            done.eq(1),
            If(enable & ~enable_d,
                NextState("FLUSH")
            ),
            sink.ready.eq(1),
            mem.source.connect(cdc.sink)
        )
        fsm.act("FLUSH",
            sink.ready.eq(1),
            mem_flush.wait.eq(1),
            mem.source.ready.eq(1),
            If(mem_flush.done,
                NextState("WAIT")
            )
        )
        fsm.act("WAIT",
            sink.connect(mem.sink, omit={"hit"}),
            If(sink.valid & sink.hit,
                NextState("RUN")
            ),
            mem.source.ready.eq(mem.level >= offset)
        )
        fsm.act("RUN",
            sink.connect(mem.sink, omit={"hit"}),
            If(mem.level >= length,
                NextState("IDLE"),
            )
        )

        # Memory read
        self.comb += [
            self.mem_valid.status.eq(cdc.source.valid),
            cdc.source.ready.eq(self.mem_data.we | ~self.enable.storage),
            self.mem_data.status.eq(cdc.source.data)
        ]


class LiteScopeAnalyzer(Module, AutoCSR):
    def __init__(self, groups, depth, clock_domain="sys", trigger_depth=16, register=False, csr_csv="analyzer.csv"):
        self.groups = groups = self.format_groups(groups)
        self.depth  = depth

        self.data_width = data_width = max([sum([len(s) for s in g]) for g in groups.values()])

        self.csr_csv = csr_csv

        # # #

        # Create scope clock domain
        self.clock_domains.cd_scope = ClockDomain()
        self.comb += self.cd_scope.clk.eq(ClockSignal(clock_domain))

        # Mux
        self.submodules.mux = _Mux(data_width, len(groups))
        sd = getattr(self.sync, clock_domain)
        for i, signals in groups.items():
            s = Cat(signals)
            if register:
                s_d = Signal(len(s))
                sd += s_d.eq(s)
                s = s_d
            self.comb += [
                self.mux.sinks[i].valid.eq(1),
                self.mux.sinks[i].data.eq(s)
            ]

        # Frontend
        self.submodules.trigger = _Trigger(data_width, depth=trigger_depth)
        self.submodules.subsampler = _SubSampler(data_width)

        # Storage
        self.submodules.storage = _Storage(data_width, depth)

        # Pipeline
        self.submodules.pipeline = stream.Pipeline(
            self.mux.source,
            self.trigger,
            self.subsampler,
            self.storage.sink)

    def format_groups(self, groups):
        if not isinstance(groups, dict):
            groups = {0 : groups}
        new_groups = {}
        for n, signals in groups.items():
            if not isinstance(signals, list):
                signals = [signals]

            split_signals = []
            for s in signals:
                if isinstance(s, Record):
                    split_signals.extend(s.flatten())
                elif isinstance(s, FSM):
                    s.do_finalize()
                    s.finalized = True
                    split_signals.append(s.state)
                else:
                    split_signals.append(s)
            new_groups[n] = split_signals
        return new_groups

    def export_csv(self, vns, filename):
        def format_line(*args):
            return ",".join(args) + "\n"
        r = format_line("config", "None", "data_width", str(self.data_width))
        r += format_line("config", "None", "depth", str(self.depth))
        for i, signals in self.groups.items():
            for s in signals:
                r += format_line("signal", str(i), vns.get_name(s), str(len(s)))
        write_to_file(filename, r)

    def do_exit(self, vns):
        if self.csr_csv is not None:
            self.export_csv(vns, self.csr_csv)



#
# This file is part of LiteScope.
#
# Copyright (c) 2015 Florent Kermarrec <florent@enjoy-digital.fr>
# SPDX-License-Identifier: BSD-2-Clause

from litescope.software.dump.common import Dump, dec2bin


class CSVDump(Dump):
    def __init__(self, dump=None):
        Dump.__init__(self)
        self.variables = [] if dump is None else dump.variables

    def generate_vars(self):
        r = ""
        for variable in self.variables:
            r += variable.name
            r += ","
        r += "\n"
        for variable in self.variables:
            r += str(variable.width)
            r += ","
        r += "\n"
        return r

    def generate_dumpvars(self):
        r  = ""
        for i in range(len(self)):
            for variable in self.variables:
                try:
                    variable.current_value = variable.values[i]
                except:
                    pass
                if variable.current_value == "x":
                    r += "x"
                else:
                    r += dec2bin(variable.current_value, variable.width)
                r += ", "
            r += "\n"
        return r

    def write(self, filename):
        f = open(filename, "w")
        f.write(self.generate_vars())
        f.write(self.generate_dumpvars())
        f.close()

    def read(self, filename):
        raise NotImplementedError("CSV files can not (yet) be read, please contribute!")



#
# This file is part of LiteScope.
#
# Copyright (c) 2015 Florent Kermarrec <florent@enjoy-digital.fr>
# SPDX-License-Identifier: BSD-2-Clause

from litescope.software.dump.common import Dump


class PythonDump(Dump):
    def __init__(self, dump=None):
        Dump.__init__(self)
        self.variables = [] if dump is None else dump.variables

    def generate_data(self):
        r = "dump = {\n"
        for variable in self.variables:
            r += "\"" + variable.name + "\""
            r += " : "
            r += str(variable.values)
            r += ",\n"
        r += "}"
        return r

    def write(self, filename):
        f = open(filename, "w")
        f.write(self.generate_data())
        f.close()

    def read(self, filename):
        raise NotImplementedError("Python files can not (yet) be read, please contribute!")



#
# This file is part of LiteScope.
#
# Copyright (c) 2015-2018 Florent Kermarrec <florent@enjoy-digital.fr>
# Copyright (c) 2019 kees.jongenburger <kees.jongenburger@gmail.com>
# Copyright (c) 2018 Sean Cross <sean@xobs.io>
# SPDX-License-Identifier: BSD-2-Clause

import os
import sys
import re

from migen import *

from litescope.software.dump.common import *
from litescope.software.dump import *

import csv


class LiteScopeAnalyzerDriver:
    def __init__(self, regs, name, config_csv=None, debug=False):
        self.regs = regs
        self.name = name
        self.config_csv = config_csv
        if self.config_csv is None:
            self.config_csv = name + ".csv"
        self.debug = debug
        self.get_config()
        self.get_layouts()
        self.build()
        self.group = 0
        self.data = DumpData(self.data_width)

        self.offset = 0
        self.length = None

        # Disable trigger and storage
        self.trigger_enable.write(0)
        self.storage_enable.write(0)

    def get_config(self):
        csv_reader = csv.reader(open(self.config_csv), delimiter=',', quotechar='#')
        for item in csv_reader:
            t, g, n, v = item
            if t == "config":
                setattr(self, n, int(v))

    def get_layouts(self):
        self.layouts = {}
        csv_reader = csv.reader(open(self.config_csv), delimiter=',', quotechar='#')
        for item in csv_reader:
            t, g, n, v = item
            if t == "signal":
                try:
                    self.layouts[int(g)].append((n, int(v)))
                except:
                    self.layouts[int(g)] = [(n, int(v))]

    def build(self):
        for key, value in self.regs.d.items():
            if self.name == key[:len(self.name)]:
                key = key.replace(self.name + "_", "")
                setattr(self, key, value)
        for signals in self.layouts.values():
            value = 1
            for name, length in signals:
                setattr(self, name + "_o", value)
                value = value*(2**length)
        for signals in self.layouts.values():
            value = 0
            for name, length in signals:
                setattr(self, name + "_m", (2**length-1) << value)
                value += length

    def configure_group(self, value):
        self.group = value
        self.mux_value.write(value)

    def add_trigger(self, value=0, mask=0, cond=None):
        if self.trigger_mem_full.read():
            raise ValueError("Trigger memory full, too much conditions")
        if cond is not None:
            for k, v in cond.items():
                # Check for binary/hexa expressions
                mb = re.match("0b([01x]+)",  v)
                mx = re.match("0x([0-fx]+)", v)
                m  = mb or mx
                if m is not None:
                    b = m.group(1)
                    v = 0
                    m = 0
                    for c in b:
                        v <<= 4 if mx is not None else 1
                        m <<= 4 if mx is not None else 1
                        if c != "x":
                            v |= int(c, 16 if mx is not None else 2 )
                            m |= 0xf if mx is not None else 0b1
                    value |= getattr(self, k + "_o")*v
                    mask  |= getattr(self, k + "_m") & (getattr(self, k + "_o")*m)
                # Else convert to int
                else:
                    value |= getattr(self, k + "_o")*int(v, 0)
                    mask  |= getattr(self, k + "_m")
        self.trigger_mem_mask.write(mask)
        self.trigger_mem_value.write(value)
        self.trigger_mem_write.write(1)

    def add_rising_edge_trigger(self, name):
        self.add_trigger(getattr(self, name + "_o")*0, getattr(self, name + "_m"))
        self.add_trigger(getattr(self, name + "_o")*1, getattr(self, name + "_m"))

    def add_falling_edge_trigger(self, name):
        self.add_trigger(getattr(self, name + "_o")*1, getattr(self, name + "_m"))
        self.add_trigger(getattr(self, name + "_o")*0, getattr(self, name + "_m"))

    def configure_trigger(self, value=0, mask=0, cond=None):
        self.add_trigger(value, mask, cond)

    def configure_subsampler(self, value):
        self.subsampler_value.write(value-1)

    def run(self, offset=0, length=None):
        if length is None:
            length = self.depth
        assert offset < self.depth
        assert length <= self.depth
        self.offset = offset
        self.length = length
        if self.debug:
            print("[running]...")
        self.storage_offset.write(offset)
        self.storage_length.write(length)
        self.storage_enable.write(1)
        self.trigger_enable.write(1)

    def clear(self):
        self.data = DumpData(self.data_width)
        self.offset = 0
        self.length = None
        self.trigger_enable.write(0)
        self.storage_enable.write(0)

    def done(self):
        return self.storage_done.read()

    def wait_done(self):
        while not self.done():
            pass

    def upload(self):
        if self.debug:
            print("[uploading]...")
        length = self.storage_length.read()
        for position in range(1, length + 1):
            if self.debug:
                sys.stdout.write("[{}>{}] {}%\r".format('=' * (20*position//length),
                                                        ' ' * (20-20*position//length),
                                                        100*position//length))
                sys.stdout.flush()
            if not self.storage_mem_valid.read():
                break
            self.data.append(self.storage_mem_data.read())
        if self.debug:
            print("")
        return self.data

    def save(self, filename, samplerate=None, flatten=False):
        if self.debug:
            print("[writing to " + filename + "]...")
        name, ext = os.path.splitext(filename)
        if ext == ".vcd":
            dump = VCDDump()
        elif ext == ".csv":
            dump = CSVDump()
        elif ext == ".py":
            dump = PythonDump()
        elif ext == ".json":
            dump = JSONDump()
        elif ext == ".sr":
            dump = SigrokDump(samplerate=samplerate)
        else:
            raise NotImplementedError
        if not flatten:
            dump.add_from_layout(self.layouts[self.group], self.data)
        else:
            dump.add_from_layout_flatten(self.layouts[self.group], self.data)
        dump.add_scope_clk()
        dump.add_scope_trig(self.offset)
        dump.write(filename)

    def get_instant_value(self, group, name):
        self.data = DumpData(self.data_width)
        self.debug = False
        self.configure_group(group)
        self.configure_trigger()
        self.configure_subsampler(1)
        self.run(0, 1)
        self.wait_done()
        self.upload()
        min_idx = log2_int(getattr(self, name + "_o"))
        max_idx = min_idx + log2_int((getattr(self, name + "_m") >> min_idx) + 1)
        return self.data[min_idx:max_idx][0]



#!/usr/bin/env python3

#
# This file is part of LiteScope.
#
# Copyright (c) 2020 Antmicro <www.antmicro.com>
# Copyright (c) 2020 Florent Kermarrec <florent@enjoy-digital.fr>
# SPDX-License-Identifier: BSD-2-Clause

import os
import re
import csv
import sys
import argparse

from litex import RemoteClient
from litescope import LiteScopeAnalyzerDriver

# Helpers ------------------------------------------------------------------------------------------

def get_signals(csvname, group):
    signals = []
    with open(csvname) as f:
        reader = csv.reader(f, delimiter=",", quotechar="#")
        for t, g, n, v in reader:
            if t == "signal" and g == group:
                signals.append(n)
    return signals

class Finder:
    def __init__(self, signals):
        self.signals = signals

    def __getitem__(self, name):
        scores = {s: 0 for s in self.signals}
        # Exact match
        if name in scores:
            print("Exact:", name)
            return name
        # Substring
        pattern = re.compile(name)
        max_score = 0
        for s in self.signals:
            match = pattern.search(s)
            if match:
                scores[s] = match.end() - match.start()
        max_score = max(scores.values())
        best = list(filter(lambda kv: kv[1] == max_score, scores.items()))
        assert len(best) == 1, f"Found multiple candidates: {best}"
        name, score = best[0]
        return name

def add_triggers(args, analyzer, signals):
    added  = False
    finder = Finder(signals)

    for signal in args.rising_edge or []:
        name = finder[signal]
        analyzer.add_rising_edge_trigger(name)
        print(f"Rising edge: {name}")
        added = True
    for signal in args.falling_edge or []:
        name = finder[signal]
        analyzer.add_falling_edge_trigger(finder[signal])
        print(f"Falling edge: {name}")
        added = True
    cond = {}
    for signal, value in args.value_trigger or []:
        name = finder[signal]
        cond[finder[signal]] = value
        print(f"Condition: {name} == {value}")
    if cond:
        analyzer.add_trigger(cond=cond)
        added = True
    return added

# Main ---------------------------------------------------------------------------------------------

def parse_args():
    parser = argparse.ArgumentParser(description="""LiteScope Client utility""")
    parser.add_argument("-r", "--rising-edge",   action="append",          help="Add rising edge trigger")
    parser.add_argument("-f", "--falling-edge",  action="append",          help="Add falling edge trigger")
    parser.add_argument("-v", "--value-trigger", action="append", nargs=2, help="Add conditional trigger with given value",
        metavar=("TRIGGER", "VALUE"))
    parser.add_argument("-l", "--list",          action="store_true",      help="List signal choices")
    parser.add_argument("--csv",                 default="analyzer.csv",   help="Analyzer CSV file")
    parser.add_argument("--csr-csv",             default="csr.csv",        help="SoC CSV file")
    parser.add_argument("--group",               default="0",              help="Capture Group")
    parser.add_argument("--subsampling",         default="1",              help="Capture Subsampling")
    parser.add_argument("--offset",              default="32",             help="Capture Offset")
    parser.add_argument("--length",              default=None,             help="Capture Length")
    parser.add_argument("--dump",                default="dump.vcd",       help="Capture Filename")
    args = parser.parse_args()
    return args

def main():
    args = parse_args()

    basename = os.path.splitext(os.path.basename(args.csv))[0]

    # Check if analyzer file is present and exit if not.
    if not os.path.exists(args.csv):
        raise ValueError("{} not found. This is necessary to load the wires which have been tapped to scope."
                         "Try setting --csv to value of the csr_csv argument to LiteScopeAnalyzer in the SoC.".format(args.csv))
        sys.exit(1)

    # Get list of signals from analyzer configuratio file.
    signals = get_signals(args.csv, args.group)

    # If in list mode, list signals and exit.
    if args.list:
        for signal in signals:
            print(signal)
        sys.exit(0)

    # Create and open remote control.
    if not os.path.exists(args.csr_csv):
        raise ValueError("{} not found. This is necessary to load the 'regs' of the remote. Try setting --csr-csv here to "
                         "the path to the --csr-csv argument of the SoC build.".format(args.csr_csv))
    bus = RemoteClient(csr_csv=args.csr_csv)
    bus.open()

    # Configure and run LiteScope analyzer.
    try:
        analyzer = LiteScopeAnalyzerDriver(bus.regs, basename, debug=True)
        analyzer.configure_group(int(args.group, 0))
        analyzer.configure_subsampler(int(args.subsampling, 0))
        if not add_triggers(args, analyzer, signals):
            print("No trigger, immediate capture.")
        analyzer.run(
            offset = int(args.offset, 0),
            length = None if args.length is None else int(args.length, 0)
        )
        analyzer.wait_done()
        analyzer.upload()
        analyzer.save(args.dump)

    # Close remove control.
    finally:
        bus.close()

if __name__ == "__main__":
    main()



#
# This file is part of LiteScope.
#
# Copyright (c) 2015-2018 Florent Kermarrec <florent@enjoy-digital.fr>
# SPDX-License-Identifier: BSD-2-Clause

from itertools import count
import datetime
from litescope.software.dump.common import Dump, dec2bin


def vcd_codes():
    codechars = [chr(i) for i in range(33, 127)]
    for n in count():
        q, r = divmod(n, len(codechars))
        code = codechars[r]
        while q > 0:
            q, r = divmod(q, len(codechars))
            code = codechars[r] + code
        yield code


class VCDDump(Dump):
    def __init__(self, dump=None, timescale="1ps", comment=""):
        Dump.__init__(self)
        self.variables = [] if dump is None else dump.variables
        self.timescale = timescale
        self.comment = comment
        self.cnt = -1

    def change(self):
        r = ""
        c = ""
        for v in self.variables:
            try:
                if v.values[self.cnt + 1] != v.current_value:
                    c += "b"
                    c += dec2bin(v.values[self.cnt + 1], v.width)
                    c += " "
                    c += v.code
                    c += "\n"
            except:
                pass
        if c != "":
            r += "#"
            r += str(self.cnt+1)
            r += "\n"
            r += c
        return r

    def generate_date(self):
        now = datetime.datetime.now()
        r = "$date\n"
        r += "\t"
        r += now.strftime("%Y-%m-%d %H:%M")
        r += "\n"
        r += "$end\n"
        return r

    def generate_version(self):
        r = "$version\n"
        r += "\tlitescope VCD dump\n"
        r += "$end\n"
        return r

    def generate_timescale(self):
        r = "$timescale "
        r += self.timescale
        r += " $end\n"
        return r

    def generate_vars(self):
        r = "$scope "
        r += self.timescale
        r += " $end\n"
        for v in self.variables:
            r += "$var wire "
            r += str(v.width)
            r += " "
            r += v.code
            r += " "
            r += v.name
            r += " $end\n"
        r += "$unscope "
        r += " $end\n"
        r += "$enddefinitions "
        r += " $end\n"
        return r

    def generate_dumpvars(self):
        r = "$dumpvars\n"
        for v in self.variables:
            v.current_value = "x"
            r += "b"
            r += dec2bin(v.current_value, v.width)
            r += " "
            r += v.code
            r += "\n"
        r += "$end\n"
        return r

    def generate_valuechange(self):
        r = ""
        for i in range(len(self)):
            r += self.change()
            self.cnt += 1
        return r

    def __repr__(self):
        r = ""
        return r

    def finalize(self):
        codegen = vcd_codes()
        for v in self.variables:
            v.code = next(codegen)

    def write(self, filename):
        self.finalize()
        f = open(filename, "w")
        f.write(self.generate_date())
        f.write(self.generate_timescale())
        f.write(self.generate_vars())
        f.write(self.generate_dumpvars())
        f.write(self.generate_valuechange())
        f.close()

    def read(self, filename):
        raise NotImplementedError("VCD files can not (yet) be read, please contribute!")



#
# This file is part of LiteScope.
#
# Copyright (c) 2015-2017 Florent Kermarrec <florent@enjoy-digital.fr>
# SPDX-License-Identifier: BSD-2-Clause

class LiteScopeIODriver:
    def __init__(self, regs, name):
        self.regs = regs
        self.name = name
        self.build()

    def build(self):
        self.input = getattr(self.regs, self.name + "_in")
        self.output = getattr(self.regs, self.name + "_out")

    def write(self, value):
        self.output.write(value)

    def read(self):
        return self.input.read()
